class _Options:
    def __init__(self):
        self.ERROR_WHEN_UNCOVERED_CODE = False
        """bool. If True, throw an AssertionError when an uncovered portion of code is reached."""


OPTIONS = _Options()
